package Lock;
use strict;
use Baseobject;
use Exporter;
use Carp ;

our @ISA=('Baseobject' );
our @EXPORT = qw (verbosity);

my $version ='$Id: Lock.pm 8602 2005-09-06 11:22:56Z anton $ ';
our $VERSION="1.1";

use Fcntl qw(:DEFAULT :flock);
use FileHandle;

sub new
{
  my $pkg=shift;
  my $self= bless (new Baseobject(@_),$pkg);
  $self->field('name','Lock');

  $self->needs('filename');

  $self->allows(
		{'trials'=>'15',
		 'sleep'=>'1',
		 'test'=>0,
		 'locked'=>0,
		 'unlink'=>0,
		 'lockpath'=>"",
		 'concat'=>'_lock',
		 'lockonly'=>0,
		 'openmode'=>"",
		 'umask'   =>umask,
		 'locktype'=>'LOCK_SH',
		 'keeplocked'=>0,
	       });

  $self->allows(['lockfilehandle','lockfilename','filehandle']);

  @_=$self->default('filename',@_) if (@_);

  $self->fields(@_);

  return $self;
}


sub open
{
#  LOCK_SH=1
#  LOCK_EX=2;
#  LOCK_UN=8;
#
#  LOCK_NB=4 ORred with
#
  my $lockmode=LOCK_SH;
  my $self=shift;


  @_=$self->default('filename',@_);
  $self->fields(@_);
  $self->checkneeds();

  return 0 if ($self->error());

  my $filename=$self->field('filename');
  my $lockfilehandle=new FileHandle();
  my $lockfilename=$filename;

  # remove > and take note if the were here
  $lockfilename =~ s/\<//g;

  #for historical (BAD) reasons, accept 'LOCK_EX' and LOCK_EX
  if ( $lockfilename =~ s/\>//g or ($self->{fields}->{locktype} eq 'LOCK_EX') or ($self->{fields}->{locktype} eq LOCK_EX) )
  {
    $lockmode=LOCK_EX;
  }

  $lockfilename.=$self->field('concat');
  if ($self->field('lockpath'))
  {
    #remove path part
    $lockfilename =~ s/^.*\///;
    $lockfilename=$self->field('lockpath')."/$lockfilename";
  }
  $self->field('lockfilehandle',$lockfilehandle);
  $self->field('lockfilename',$lockfilename);
  #
  # if filename contains at least one '>'
  # exclusive lock mode is enabled
  # else shared lock is sufficient
  #

  my $modes=
  {
    1 => 'LOCK_SH',
    2 => 'LOCK_EX',
    8 => 'LOCK_UN',
  };


  my $mask=umask;
  umask $self->field('umask');

  unless (open ($lockfilehandle,"+>$lockfilename"))
  {
    umask $mask;
    $self->stderror( "Could not open: $lockfilename, $!");
    return 0;
  }

  my $trials=$self->field('trials');

  if ($self->field('locked') == $lockmode)
  {
    $self->debug("already locked, lockmode: $lockmode");
    return 0 ;
  }

  while (! (flock($lockfilehandle, $lockmode|LOCK_NB)) && $trials)
  {
    warn "Waiting for lock on: $lockfilename: $trials seconds to go: $!\n" if (verbosity()>=2);
    $trials--;
    sleep $self->field('sleep');
    unless ($trials)
    {
      umask $mask;
      close $lockfilehandle;
      $self->stderror("timeout occurred for $lockfilename: $!\n");
      return 0;
    }
  }

  $self->field('locked',$lockmode);
  $self->debug( "locked ($modes->{$lockmode}): $lockfilename\n");

  my $fh=-1;
  unless ($self->field('lockonly'))
  {
    $fh= new FileHandle();
    unless (open ($fh,$filename))
    {
      $self->error("Could not open file: \"$filename\": $!\n") ;
      umask $mask;
      return 0;
    }

    $self->field('filehandle',$fh);
    warn "opened: $filename\n" if (verbosity()>=2);
  }
  else
  {
    warn "locked only" if (verbosity()>=2);
  }

  if ($self->field('test'))
  {
    warn "press enter to continue";
    <>;
  }
  umask $mask;
  return $fh;
}



sub close
{
  my $self=shift;
  #unlock
  my $lockfilehandle=$self->field('lockfilehandle');
  my $lockfilename=$self->field('lockfilename');

  return unless (defined ($lockfilehandle));

  if ($self->field('keeplocked'))
  {
#    warn "KEEPLOCKED" ;
    return ;
  }

  if ($self->field('unlink'))
  {
    unlink $lockfilename or
	warn  "Could not unlink file: lockfilename";

  }

  flock($lockfilehandle,LOCK_UN) or warn "Could not unlock ";
  CORE::close($lockfilehandle);

  $self->field('locked',0);
  $self->debug( "unlocked $lockfilename\n");


  unless ($self->field('lockonly'))
  {
    my $filehandle=$self->field('filehandle');
    my $filename=$self->field('filename');

    CORE::close($filehandle);
    warn "close $filename\n" if (verbosity()>=2);
  }

}

sub DESTROY
{
  my $self=shift;
  $self->close() if $self->field('locked');
}


sub default
{
  # first version of a sequence interpretor
  my $self=shift;
  my $param=shift;
  my $extra = $_[0];
  #
  # default 1 parameter: and it's a scalar
  #

  if (defined $_[0] == 1)
  {
     $self->field ($param,shift) unless (ref ($_[0]));
  }
  @_;
}

42;

=head1 Lock.pm

=over 1

=head1 SYNOPSIS

Create an open filehandle and lock it at the same time.

=head1 DESCRIPTION

A 'trials' number of seconds an open attempt on 'filename' is
made. During the period the file is opened a volontary lockmechanisme
is placed around the operation.

Next to the data file an additional lockfile: 'filename'.'concat' is
created for the time being and optionally (default) removed after
usage.

=head1 DEPENDENCY

Inherits from Baseobject.

=head1 EXPORTS

verbosity (see Baseobject)

=head1 EXAMPLES

=head2 EXAMPLE1

 use Lock;

 my $lock= new Lock();
 my $fh=$lock->open('>>myfile');

 die $lock->error() unless ($fh) ;
 print $fh "just testing\n";
 $lock->close();

=head2 EXAMPLE2

 use Lock;
 verbosity(2);

 my $lock= new Lock(
           {
             #open this file for appending
             'filename'=>'>>myfile',
             #try 5 times
             'trials'=>5,
             #do not remove the lock file after usage
             'unlink'=>0,
             #place lockfile on /tmp
             'lockpath'=>"/tmp",
             'concat'=>"__mylock",
             #wait for stdin before releasing the lock
             'test'=>1,
           } );

 my $fh=$lock->open();
 die $lock->error()  unless ($fh) ;
 print $fh "just testing\n";
 $lock->close();

=head1 METHODS

=head2 open ()

=head1 SYNOPSIS

Opens locked file.

=head1 PARAMETERS

same as L<new ()|new ()>

=head1 RETURNS

Returns a filehandle unless 'lockonly' is 1.

=head2 new ()

=item SYNOPSIS

Create a new lock-object.

=item PARAMETERS

either a scalar, which is interpreted as B<filename>, or
a ref to a hash containing the following elements:


<filename>
Filename passed to a perl open statment. To open a file for writing use the '>' or '>>' arrows.

I<lockpath>
Use this directory to place the lockfile in stead of the location of the datafile. The name of the file is allways the above filename concatenated with concat

    concat
        The string concattenated after the filename. Default: '_lock'

    trials
        Number of trials before failure.

        default: 5

    sleep
        Sleep interval between the trials in seconds.

        default: 1

    unlink
        The corresponding 'filename'.'concat' file is removed if set to 1.

        default:0

    test
        If set to true, the process waits carriage return on stdin. For testing purposes, obviously.

    lockonly
        Do not open a file and return -1 as filehandle. Only the lockfile is created and locked.

    default: 0

other fields:
    inherited from Baseobject.







